/*
 * Copyright (c) 1985, 1988, 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "../main/dd.h"

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>

#define	FTP_NAMES
#include <ftp.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>

#include <stdarg.h>
#include <stdio.h>
#include <grp.h>
#include <signal.h>
#include <dirent.h>
#include <fcntl.h>
#include <time.h>
#ifdef SHADOW_PWD
#include <shadow.h>
#endif
#ifdef HAVE_UTMPX_H
#include <utmpx.h>
#endif
#include <pwd.h>
#include <setjmp.h>
#include <netdb.h>
#include <errno.h>
#include <syslog.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "pathnames.h"

#include <global.h>
#include <md5.h>
#include <ddftp.h>
#include <ddcommon.h>

#define MD_CTX MD5_CTX
#define MDInit MD5Init
#define MDUpdate MD5Update
#define MDFinal MD5Final

/*
 * File containing login names
 * NOT to be used on this machine.
 * Commonly used to disallow uucp.
 */
extern	int errno;
extern	char version[];
extern	char *home;		/* pointer to home directory for glob */
extern	FILE *ftpd_popen(), *fopen(), *freopen();
extern	int  ftpd_pclose(), fclose();
extern	char *getline();
extern	char cbuf[];
extern	off_t restart_point;

struct	sockaddr_in ctrl_addr;
struct	sockaddr_in data_source;
struct	sockaddr_in data_dest;
struct	sockaddr_in his_addr;
struct	sockaddr_in pasv_addr;

int	data;
jmp_buf	errcatch, urgcatch;
int	logged_in;
struct	userbase *pw;
int	debug;
int	timeout = 1800;    /* timeout after 15 minutes of inactivity */
int	maxtimeout = 7200;/* don't allow idle time to be set beyond 2 hours */
int	logging;
int	restricted_data_ports = 1;
int	guest;
int	type;
int	form;
int	stru;			/* avoid C keyword */
int	mode;
int	usedefault = 1;		/* for data transfers */
int	pdata = -1;		/* for passive mode */
int     ulconf = 0;
int	transflag;
off_t	file_size;
off_t	byte_count;
#if !defined(CMASK) || CMASK == 0
#undef CMASK
#define CMASK 027
#endif
int	defumask = CMASK;		/* default umask value */
char	tmpline[7];
char	hostname[MAXHOSTNAMELEN];
char	remotehost[MAXHOSTNAMELEN];


/*
 * Timeout intervals for retrying connections
 * to hosts that don't accept PORT cmds.  This
 * is a kludge, but given the problems with TCP...
 */
#define	SWAITMAX	90	/* wait at most 90 seconds */
#define	SWAITINT	5	/* interval between retries */

int	swaitmax = SWAITMAX;
int	swaitint = SWAITINT;

void	lostconn(), myoob();
FILE	*getdatasock(), *dataconn();

#ifndef HAVE_SETPROCTITLE
char	**Argv = NULL;		/* pointer to argument vector */
char	*LastArgv = NULL;	/* end of argv */
#endif /* HAVE_SETPROCTITLE */
char	proctitle[BUFSIZ];	/* initial part of title */

char datadir[1024];
struct DayDream_MainConfig mcfg;
struct DayDream_Conference *confs=0;

int mydir=0;
char myroot[1024];
static void end_login(void);
static char *fgetsnolf(char *buf, int n, FILE *fh);
static int findfile(const char *file, char *de, struct DayDream_Conference *tc);
static void _splitpath(char *file, char *path, char *name);
static int cmppasswds(char *passwd, unsigned char *thepw);
static int dupecheck(const char *name);
static int receive_data(FILE *instr, FILE *outstr);
static void send_data(FILE *instr, FILE *outstr, off_t blksize);
static int checkpath(char *pa);

static struct sockaddr_in passive_ip;
static int forced_passive_ip = 0;

struct ftpinfo minfo;
char *infoname=0;
static char *daydream_path = NULL;

int getddconfig(void)
{
	char buf[1024];
	int fd;
	struct stat fib;
	unsigned char *s;
	
	snprintf(buf, sizeof(buf), "%sdata/daydream.dat", datadir);
	if ((fd = open(buf, O_RDONLY)) == -1)
		return -1;
	if (read(fd, &mcfg, sizeof(struct DayDream_MainConfig)) !=
		sizeof(struct DayDream_MainConfig))
		return -1;
	close(fd);

	snprintf(buf, sizeof(buf), "%sdata/conferences.dat", datadir);
	if ((fd = open(buf,O_RDONLY)) == -1)
		return -1;
	if (fstat(fd,&fib) == -1)
		return -1;
	if (fib.st_size % sizeof(struct DayDream_Conference))
		return -1;
	confs = (struct DayDream_Conference *) malloc(fib.st_size + 2);
	if (confs == NULL)
		return -1;
	if (read(fd, confs, fib.st_size) != fib.st_size)
		return -1;
	close(fd);
	s=(unsigned char *)confs;
	s[fib.st_size]=255;

	memset(&minfo,0,sizeof(struct ftpinfo));
	return 0;
}

void showfile(const char *foo)
{
	FILE *fi;
	char buf[1024];
	char *s;
	sprintf(buf,"%sdisplay/%s",datadir,foo);
	fi=fopen(buf,"r");
	if (fi) {
		while(fgets(buf,1024,fi)) {
                    if ((s = strchr(buf, '\n')) != NULL)
                        *s = '\0';
		    lreply(250,buf);
		}
		lreply(250,"");
		fclose(fi);
	}
}

static int doinitgroups(uid_t uid, gid_t gid)
{
	struct passwd *pwd;

	if ((pwd = getpwuid(uid)) == NULL)
		return -1;
	return initgroups(pwd->pw_name, gid);
}

int main(int argc, char *argv[], char *envp[])
{
	int addrlen, on = 1, tos;
	char *cp;

	init_common_library(argc, argv);

	/*
	 * LOG_NDELAY sets up the logging connection immediately,
	 * necessary for anonymous ftp's that chroot and can't do it later.
	 */
	openlog("ftpd", LOG_PID | LOG_NDELAY, LOG_DAEMON);
	addrlen = sizeof (his_addr);
	if (getpeername(0, (struct sockaddr *)&his_addr, &addrlen) < 0) {
		syslog(LOG_ERR, "getpeername (%s): %m",argv[0]);
		exit(1);
	}
	addrlen = sizeof (ctrl_addr);
	if (getsockname(0, (struct sockaddr *)&ctrl_addr, &addrlen) < 0) {
		syslog(LOG_ERR, "getsockname (%s): %m",argv[0]);
		exit(1);
	}
#ifdef IP_TOS
	tos = IPTOS_LOWDELAY;
	if (setsockopt(0, IPPROTO_IP, IP_TOS, (char *)&tos, sizeof(int)) < 0)
		syslog(LOG_WARNING, "setsockopt (IP_TOS): %m");
#endif
	data_source.sin_port = htons(ntohs(ctrl_addr.sin_port) - 1);
	debug = 0;
#ifndef HAVE_SETPROCTITLE
	/*
	 *  Save start and extent of argv for setproctitle.
	 */
	Argv = argv;
	while (*envp)
		envp++;
	LastArgv = envp[-1] + strlen(envp[-1]);
#endif /* HAVE_SETPROCTITLE */

	argc--, argv++;
	while (argc > 0 && *argv[0] == '-') {
		for (cp = &argv[0][1]; *cp; cp++) switch (*cp) {

		case 'v':
			debug = 1;
			break;

		case 'd':
			debug = 1;
			break;

		case 'l':
			logging = 1;
			break;

		case 'U':
			restricted_data_ports = 0;
			break;

		case 't':
			timeout = atoi(++cp);
			if (maxtimeout < timeout)
				maxtimeout = timeout;
			goto nextopt;

		case 'T':
			maxtimeout = atoi(++cp);
			if (timeout > maxtimeout)
				timeout = maxtimeout;
			goto nextopt;

		case 'u':
		    {
			int val = 0;

			while (*++cp && *cp >= '0' && *cp <= '9')
				val = val*8 + *cp - '0';
			if (*cp)
				fprintf(stderr, "ftpd: Bad value for -u\n");
			else
				defumask = val;
			goto nextopt;
		    }
		case 'D':
			/* FIXME: buffer overflow */
			strlcpy(datadir, ++cp, sizeof(datadir));
			if (getddconfig() == -1) {
				fprintf(stderr, "ftpd: cannot load config\n");
				exit(1);
			}
			setenv("DAYDREAM", datadir, 1);
			goto nextopt;
		case 'P':
			if (inet_pton(AF_INET, ++cp, 
				&passive_ip.sin_addr) <= 0) {
				fputs("ftpd: invalid forced passive IP\n",
					stderr);
				exit(1);
			}
			forced_passive_ip = 1;
			goto nextopt;
		case 'p':
			if (getenv("PATH") == NULL)
				setenv("PATH", strdup(++cp), 1);
			else {
				cp++;
				daydream_path = (char *) malloc(strlen(cp) +
					strlen(getenv("PATH")) + 2);
				sprintf(daydream_path, "%s:%s", 
					getenv("PATH"), cp);
				setenv("PATH", daydream_path, 1);
			}
			goto nextopt;
			
		default:
			fprintf(stderr, "ftpd: Unknown flag -%c ignored.\n",
			     *cp);
			break;
		}
nextopt:
		argc--, argv++;
	}
	if (!confs) {
		fprintf(stderr,"No config loaded!\n");
		exit(1);
	}

	if (doinitgroups(mcfg.CFG_BBSUID, mcfg.CFG_BBSGID) == -1) {
		syslog(LOG_ERR, "initgroups() failed");
		exit(1);
	}

	(void) freopen(_PATH_DEVNULL, "w", stderr);
	(void) signal(SIGPIPE, lostconn);
	(void) signal(SIGCHLD, SIG_IGN);
	if ((int)signal(SIGURG, myoob) < 0)
		syslog(LOG_ERR, "signal: %m");

	/* Try to handle urgent data inline */
#ifdef SO_OOBINLINE
	if (setsockopt(0, SOL_SOCKET, SO_OOBINLINE, (char *)&on, sizeof(on)) < 0)
		syslog(LOG_ERR, "setsockopt: %m");
#endif

#ifdef	F_SETOWN
	if (fcntl(fileno(stdin), F_SETOWN, getpid()) == -1)
		syslog(LOG_ERR, "fcntl F_SETOWN: %m");
#endif
	dolog(&his_addr);
	/*
	 * Set up default state
	 */
	data = -1;
	type = TYPE_I;
	form = FORM_N;
	stru = STRU_F;
	mode = MODE_S;
	tmpline[0] = '\0';
	(void) gethostname(hostname, sizeof (hostname));
	reply(220, "%s DD-FTP server (%s) ready.", hostname, version);
	(void) setjmp(errcatch);
	for (;;)
		(void) yyparse();
	/* NOTREACHED */
}

#ifndef HAVE_SETPROCTITLE
/*
 * clobber argv so ps will show what we're doing.
 * (stolen from sendmail)
 * warning, since this is usually started from inetd.conf, it
 * often doesn't have much of an environment or arglist to overwrite.
 */

/*VARARGS1*/
void setproctitle(fmt, a, b, c)
char *fmt;
{
	register char *p, *bp, ch;
	register int i;
	char buf[BUFSIZ];

	(void) sprintf(buf, fmt, a, b, c);

	/* make ps print our process name */
	p = Argv[0];
	*p++ = '-';

	i = strlen(buf);
	if (i > LastArgv - p - 2) {
		i = LastArgv - p - 2;
		buf[i] = '\0';
	}
	bp = buf;
	while ((ch = *bp++) != 0)
		if (ch != '\n' && ch != '\r')
			*p++ = ch;
	while (p < LastArgv)
		*p++ = ' ';
}
#endif /* HAVE_SETPROCTITLE */

void
lostconn()
{
	if (debug)
		syslog(LOG_DEBUG, "lost connection");
	dologout(-1);
}

static char ttyline[20];

static char *filepart(char *s)
{
	char *t;

	t = s + strlen(s); 
	while(s != t) {
		if (*t == '/') 
			return t + 1;
		t--;
	}
	return t;
}

/*
 * Save the result of a getpwnam.  Used for USER command, since
 * the data returned must not be clobbered by any other command
 * (e.g., globbing).
 */
static struct userbase *ddgetpwnam(char *name)
{
	int fd;
	char buf[1024];
	static struct userbase ub;
	
	sprintf(buf,"%sdata/userbase.dat",datadir);
	
	fd=open(buf,O_RDONLY);
	if (fd > -1) {
		while(read(fd,&ub,sizeof(struct userbase))) {
			if ((ub.user_toggles & (1L<<30)) && ((ub.user_toggles & (1L<<31)) == 0)) continue;
			if ( !strcasecmp(ub.user_handle,name) || !strcasecmp(ub.user_realname,name)) { 
				close(fd);
				return(&ub);
			}
		}
		close(fd);
	} else 
		syslog(LOG_ERR, "cannot open userbase \"%sdata/userbase.dat\" (%m)", datadir);
	return 0;
}

int login_attempts;		/* number of failed login attempts */
int askpasswd;			/* had user command, ask for passwd */

/*
 * USER command.
 * Sets global passwd pointer pw if named account exists and is acceptable;
 * sets askpasswd if a PASS command is expected.  If logged in previously,
 * need to reset state.  If name is "ftp" or "anonymous", the name is not in
 * _PATH_FTPUSERS, and ftp account exists, set guest and pw, then just return.
 * If account doesn't exist, ask for passwd anyway.  Otherwise, check user
 * requesting login privileges.  Disallow anyone who does not have a standard
 * shell as returned by getusershell().  Disallow anyone mentioned in the file
 * _PATH_FTPUSERS to allow people such as root and uucp to be avoided.
 */
void user(char *name)
{
	char *getusershell();

	if (logged_in) {
		if (guest) {
			reply(530, "Can't change user from guest login.");
			return;
		}
		end_login();
	}

	guest = 0;

	pw = ddgetpwnam(name);
	if (!pw) {
		reply(530,"User %s unknown.",name);
		return;
	}
	reply(331, "Password required for %s.", name);
	askpasswd = 1;
	/*
	 * Delay before reading passwd after first failed
	 * attempt to slow down passwd-guessing programs.
	 */
	if (login_attempts)
		sleep((unsigned) login_attempts);
}

static int ispid(pid_t pid)
{
	return kill(pid, 0) != -1;
}

static int checkconfaccess(int confn)
{
        int newcn=confn-1;

        if (newcn < 32) {
                if (pw->user_conferenceacc1 & (1L<<newcn)) return 1;
        } else {
                newcn-=32;
                if (pw->user_conferenceacc2 & (1L<<newcn)) return 1;
        }
        return 0;
}

int setulconf(int cn)
{
	struct DayDream_Conference *tconf=confs;
	struct DayDream_MsgBase *tbase;
	int bcnt;

	if (!checkconfaccess(cn)) return 0;
	while(tconf->CONF_NUMBER!=255 && tconf->CONF_NUMBER) {
		if (tconf->CONF_UPLOADAREA && cn==tconf->CONF_NUMBER) {
			ulconf=cn;
			return 1;
		}
                /* FIXME: Very suspicious code */
		tbase = (struct DayDream_MsgBase *) tconf + 1;
		bcnt=tconf->CONF_MSGBASES;
		while(bcnt) {
			tbase++;
			bcnt--;
		}
		tconf=(struct DayDream_Conference *)tbase;
	}
	return 0;
}

char *getddinfo()
{
	static char nam[256];
	int infd;
	struct ftpinfo mi;
	int i;
	char *res=0;
	int freeh=0;
	for (i=0;i<mcfg.CFG_MAXFTPUSERS;i++) {
		sprintf(nam, "%s/ftpinfo%d.dat", DDTMP, i+1);
		infd = open(nam, O_RDONLY);
		if (infd<0) {
		kamake:
			freeh=i+1;
			res=nam;
		} else {
			read(infd,&mi,sizeof(struct ftpinfo));
			close(infd);
			if (!ispid(mi.pid)) {
				goto kamake;
			}
			if ( ((mcfg.CFG_FLAGS & (1L<<12))==0) &&   mi.userid==pw->user_account_id) return (char *)-1;
		}
	}
	if (res) {
		sprintf(nam, "%s/ftpinfo%d.dat", DDTMP, freeh);
		infd=open(nam,O_WRONLY|O_CREAT|O_TRUNC,0644);
		minfo.pid=getpid();
		minfo.userid=pw->user_account_id;
		if (infd<0) return 0;
		write(infd,&minfo,sizeof(struct ftpinfo));
		close(infd);
	}
	return res;
}

static void updateinfo(void)
{
	int infd=open(infoname,O_WRONLY|O_CREAT|O_TRUNC,0644);
	minfo.pid=getpid();
	minfo.userid=pw->user_account_id;
	if (infd<0) return;
	write(infd,&minfo,sizeof(struct ftpinfo));
	close(infd);
}
/*
 * Check if a user is in the file _PATH_FTPUSERS
 */
int checkuser(char *name)
{
	register FILE *fd;
	register char *p;
	char line[BUFSIZ];

	if ((fd = fopen(_PATH_FTPUSERS, "r")) != NULL) {
		while (fgets(line, sizeof(line), fd) != NULL)
			if ((p = strchr(line, '\n')) != NULL) {
				*p = '\0';
				if (line[0] == '#')
					continue;
				if (strcmp(line, name) == 0)
					return (1);
			}
		(void) fclose(fd);
	}
	return (0);
}

static void update_wtmp(void)
{
#ifdef HAVE_UPDWTMPX
	struct utmpx ut;
	ut.ut_user[0] = '\0';
	ut.ut_id[0] = 'f';
	ut.ut_id[1] = 't';
	ut.ut_id[2] = 'p';
	ut.ut_id[3] = '\xff';
	ut.ut_type = USER_PROCESS;
	strlcpy(ut.ut_line, ttyline, sizeof(ut.ut_line)); 
	ut.ut_pid = getpid();
#ifdef HAVE_STRUCT_UTMPX_E_EXIT
	ut.ut_exit.e_termination = 0;
	ut.ut_exit.e_exit = 0;
#endif
	ut.ut_session = 0;
	gettimeofday(&ut.ut_tv, NULL);
#ifdef HAVE_STRUCT_UTMPX_UT_SYSLEN
	ut.ut_syslen = 1;
#endif
	ut.ut_host[0] = '\0';
#else
	logwtmp(ttyline, "", "");
#endif
}

/*
 * Terminate login as previous user, if any, resetting state;
 * used when USER command is given or login fails.
 */
static void end_login(void)
{

	(void) seteuid((uid_t)0);
	if (logged_in) 
		update_wtmp();
	pw = NULL;
	logged_in = 0;
	guest = 0;
}

void pass(char *passwd)
{
	char buf[1024];
	if (logged_in || askpasswd == 0) {
		reply(503, "Login with USER first.");
		return;
	}
	askpasswd = 0;
	if (!guest) {		/* "ftp" is only account allowed no password */
		if (!cmppasswds(passwd,pw->user_password)) {
			reply(530, "Login incorrect.");
			pw = NULL;
			if (login_attempts++ >= 5) {
				syslog(LOG_NOTICE,
				    "repeated login failures from %s",
				    remotehost);
				exit(0);
			}
			return;
		}
	}
	
	login_attempts = 0;		/* this time successful */

	/* open wtmp before chroot */
	(void)sprintf(ttyline, "ftp%d", getpid());
	logged_in = 1;

	sprintf(buf,"%susers/%d/ftp",datadir,pw->user_account_id);
	mkdir(buf,0755);
	chown(buf,mcfg.CFG_BBSUID,-1);
	sprintf(buf,"%susers/%d/ftp/dl",datadir,pw->user_account_id);
	mkdir(buf,0755);
	chown(buf,mcfg.CFG_BBSUID,-1);
	sprintf(buf,"%susers/%d/ftp/ul",datadir,pw->user_account_id);
	mkdir(buf,0755);
	chown(buf,mcfg.CFG_BBSUID,-1);
	
	sprintf(buf,"%susers/%d/ftp",datadir,pw->user_account_id);
	if (chdir(buf) < 0) {
		reply(550, "Can't get user directory.");
		goto bad;
	}
	getcwd(myroot, sizeof(myroot));
	seteuid((uid_t)0);
	if (setegid((gid_t) mcfg.CFG_BBSGID) < 0) {
		reply(550, "Cannot set GID.");
		goto bad;
	}
	if (seteuid((uid_t)mcfg.CFG_BBSUID) < 0) {
		reply(550, "Cannot set UID.");
		goto bad;
	}
	infoname=getddinfo();
	if (!infoname) {
		reply(550,"No free ftp-slots. Max %d. Try again later.",mcfg.CFG_MAXFTPUSERS);
		goto bad;
	} else if ((int)infoname==-1) {
		reply(550,"No multiple ftp-connections allowed!");
		goto bad;
	}

        lreply(230, "Help");
	printf("Upload files to ul/ dir. After uploading log on to the BBS and hit U\nin the destination conference.\n\nFiles to DL are in dl/ dir. Please remove them after you've leeched them.\n\n");
	reply(230, "User %s logged in.", pw->user_handle);

	sprintf(proctitle, "%s: %s", remotehost, pw->user_handle);
	setproctitle(proctitle);

	home = datadir;		/* home dir for globbing */
	(void) umask(defumask);
	return;
bad:
	/* Forget all about it... */
	end_login();
}

void retrieve(char *cmd, char *name)
{
	FILE *fin, *dout;
	struct stat st;
	int (*closefunc)();

	if (!checkpath(name)) {
		reply(550,"Illegal filename");
		return;
	}
	strncpy(minfo.filename,filepart(name),256);
	if (cmd == 0) {
		fin = fopen(name, "r"), closefunc = fclose;
		st.st_size = 0;
		
	} else {
		char line[BUFSIZ];

		(void) sprintf(line, cmd, name), name = line;
		fin = ftpd_popen(line, "r"), closefunc = ftpd_pclose;
		st.st_size = -1;
		st.st_blksize = BUFSIZ;
	}
	if (fin == NULL) {
		if (errno != 0)
			perror_reply(550, name);
		return;
	}
	if (cmd == 0 &&
	    (fstat(fileno(fin), &st) < 0 || (st.st_mode&S_IFMT) != S_IFREG)) {
		reply(550, "%s: not a plain file.", name);
		goto done;
	}
	if (restart_point) {
		if (type == TYPE_A) {
			register int i, n, c;

			n = restart_point;
			i = 0;
			while (i++ < n) {
				if ((c=getc(fin)) == EOF) {
					perror_reply(550, name);
					goto done;
				}
				if (c == '\n')
					i++;
			}	
		} else if (lseek(fileno(fin), restart_point, SEEK_SET) < 0) {
			perror_reply(550, name);
			goto done;
		}
	}
	dout = dataconn(name, st.st_size, "w");
	if (dout == NULL)
		goto done;
	send_data(fin, dout, st.st_blksize);
	(void) fclose(dout);
	data = -1;
	pdata = -1;
done:
	(*closefunc)(fin);
}

void _store(char *name, const char *mode, int unique)
{
	FILE *fout, *din;
	struct stat st;
	char buf[1024];
	int (*closefunc)();
	char *gunique();
	
	if (!checkpath(name)) {
		reply(550,"Illegal filename");
		return;
	}

	if (dupecheck(name)) {
		reply(550,"File %s is already online.",name);
		return;
	}

	if (unique && stat(name, &st) == 0 &&
	    (name = gunique(name)) == NULL)
		return;

	if (restart_point)
		mode = "r+w";
	fout = fopen(name, mode);
	closefunc = fclose;
	if (fout == NULL) {
		perror_reply(553, name);
		return;
	}
	if (restart_point) {
		if (type == TYPE_A) {
			register int i, n, c;

			n = restart_point;
			i = 0;
			while (i++ < n) {
				if ((c=getc(fout)) == EOF) {
					perror_reply(550, name);
					goto done;
				}
				if (c == '\n')
					i++;
			}	
			/*
			 * We must do this seek to "current" position
			 * because we are changing from reading to
			 * writing.
			 */
			if (fseek(fout, 0L, SEEK_CUR) < 0) {
				perror_reply(550, name);
				goto done;
			}
		} else if (lseek(fileno(fout), restart_point, SEEK_SET) < 0) {
			perror_reply(550, name);
			goto done;
		}
	}
	din = dataconn(name, (off_t)-1, "r");
	strncpy(minfo.filename,filepart(name),256);

	if (din == NULL)
		goto done;
	if (receive_data(din, fout) == 0) {
		(void) fclose(din);
		getcwd(&buf[512],512);
		if (ulconf) {
			chdir(datadir);
			sprintf(buf, "daydream -u %d %d %s/%s",
				pw->user_account_id, ulconf,
				&buf[512], filepart(name));
			system(buf);
			chdir(&buf[512]);
		}
		if (unique)
			reply(226, "Transfer complete (unique file name:%s).",
			    name);
		else
			reply(226, "Transfer complete.");
	} else {
		(void) fclose(din);
	}
	data = -1;
	pdata = -1;
done:
	(*closefunc)(fout);
}

FILE *getdatasock(char *mode)
{
	int s, on = 1, tries;

	if (data >= 0)
		return (fdopen(data, mode));
	(void) seteuid((uid_t)0);
	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		goto bad;
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
	    (char *) &on, sizeof (on)) < 0)
		goto bad;
	/* anchor socket to avoid multi-homing problems */
	data_source.sin_family = AF_INET;
	data_source.sin_addr = ctrl_addr.sin_addr;
	for (tries = 1; ; tries++) {
		if (bind(s, (struct sockaddr *)&data_source,
		    sizeof (data_source)) >= 0)
			break;
		if (errno != EADDRINUSE || tries > 10)
			goto bad;
		sleep(tries);
	}
	(void) seteuid((uid_t)mcfg.CFG_BBSUID);
#ifdef IP_TOS
	on = IPTOS_THROUGHPUT;
	if (setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int)) < 0)
		syslog(LOG_WARNING, "setsockopt (IP_TOS): %m");
#endif
	return (fdopen(s, mode));
bad:
	(void) seteuid((uid_t)mcfg.CFG_BBSUID);
	(void) close(s);
	return (NULL);
}

FILE *dataconn(char *name, off_t size, char *mode)
{
	char sizebuf[32];
	FILE *file;
	int retry = 0, tos;

	minfo.filesize=size;
	file_size = size;
	byte_count = 0;
	if (size != (off_t) -1)
		(void) sprintf (sizebuf, " (%ld bytes)", size);
	else
		(void) strcpy(sizebuf, "");
	if (pdata >= 0) {
		struct sockaddr_in from;
		int s, fromlen = sizeof(from);

		s = accept(pdata, (struct sockaddr *)&from, &fromlen);
		if (s < 0) {
			reply(425, "Can't open data connection.");
			(void) close(pdata);
			pdata = -1;
			return(NULL);
		}
		(void) close(pdata);
		pdata = s;
#ifdef IP_TOS
		tos = IPTOS_LOWDELAY;
		(void) setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&tos,
		    sizeof(int));
#endif
		reply(150, "Opening %s mode data connection for %s%s.",
		     type == TYPE_A ? "ASCII" : "BINARY", name, sizebuf);
		return(fdopen(pdata, mode));
	}
	if (data >= 0) {
		reply(125, "Using existing data connection for %s%s.",
		    name, sizebuf);
		usedefault = 1;
		return (fdopen(data, mode));
	}
	if (usedefault)
		data_dest = his_addr;
	usedefault = 1;
	file = getdatasock(mode);
	if (file == NULL) {
		reply(425, "Can't create data socket (%s,%d): %s.",
		    inet_ntoa(data_source.sin_addr),
		    ntohs(data_source.sin_port), strerror(errno));
		return (NULL);
	}
	data = fileno(file);
	while (connect(data, (struct sockaddr *)&data_dest,
	    sizeof (data_dest)) < 0) {
		if (errno == EADDRINUSE && retry < swaitmax) {
			sleep((unsigned) swaitint);
			retry += swaitint;
			continue;
		}
		perror_reply(425, "Can't build data connection");
		(void) fclose(file);
		data = -1;
		return (NULL);
	}
	reply(150, "Opening %s mode data connection for %s%s.",
	     type == TYPE_A ? "ASCII" : "BINARY", name, sizebuf);
	return (file);
}

/*
 * Tranfer the contents of "instr" to
 * "outstr" peer using the appropriate
 * encapsulation of the data subject
 * to Mode, Structure, and Type.
 *
 * NB: Form isn't handled.
 */
static void send_data(FILE *instr, FILE *outstr, off_t blksize)
{
	register int c, cnt;
	register char *buf;
	int netfd, filefd;
	time_t startt, t2;
	startt=time(0);
	transflag++;
	if (setjmp(urgcatch)) {
		transflag = 0;
		return;
	}
	switch (type) {

	case TYPE_A:
		while ((c = getc(instr)) != EOF) {
			byte_count++;
			if (c == '\n') {
				if (ferror(outstr))
					goto data_err;
				(void) putc('\r', outstr);
			}
			(void) putc(c, outstr);
		}
		fflush(outstr);
		transflag = 0;
		if (ferror(instr))
			goto file_err;
		if (ferror(outstr))
			goto data_err;
		reply(226, "Transfer complete.");
		return;

	case TYPE_I:
	case TYPE_L:
		minfo.mode=2;
		if ((buf = malloc((u_int)blksize)) == NULL) {
			transflag = 0;
			perror_reply(451, "Local resource failure: malloc");
			return;
		}
		netfd = fileno(outstr);
		filefd = fileno(instr);
		minfo.transferred=0;
		while ((cnt = read(filefd, buf, (u_int)blksize)) > 0) {
			int offset = 0, sent;

			while (cnt > 0) {
				sent = write(netfd, buf+offset, cnt);
				minfo.transferred+=sent;
				t2=time(0);
				if (t2-startt) minfo.cps=minfo.transferred/(t2-startt);
				updateinfo();
				if (sent <= 0) break;
				offset += sent;
				cnt -= sent;
			}
			byte_count += cnt;
			if (cnt) break;
		}
		transflag = 0;
		(void)free(buf);
		minfo.mode=0;
		updateinfo();
		if (cnt != 0) {
			if (cnt < 0)
				goto file_err;
			goto data_err;
		}
		reply(226, "Transfer complete.");
		return;
	default:
		transflag = 0;
		reply(550, "Unimplemented TYPE %d in send_data", type);
		return;
	}

data_err:
	transflag = 0;
	perror_reply(426, "Data connection");
	return;

file_err:
	transflag = 0;
	perror_reply(551, "Error on input file");
}

/*
 * Transfer data from peer to
 * "outstr" using the appropriate
 * encapulation of the data subject
 * to Mode, Structure, and Type.
 *
 * N.B.: Form isn't handled.
 */
static int receive_data(FILE *instr, FILE *outstr)
{
	register int c;
	int cnt, bare_lfs = 0;
	char buf[BUFSIZ];
	time_t startt=time(0);
	time_t t2;

	transflag++;
	if (setjmp(urgcatch)) {
		transflag = 0;
		return (-1);
	}
	switch (type) {

	case TYPE_I:
	case TYPE_L:
		minfo.mode=1;
		while ((cnt = read(fileno(instr), buf, sizeof buf)) > 0) {
			if (write(fileno(outstr), buf, cnt) != cnt)
				goto file_err;
			byte_count += cnt;
			minfo.transferred=byte_count;
			t2=time(0);
			if (t2-startt) {
				minfo.cps=byte_count/(t2-startt);
			}
			updateinfo();
		}
		if (cnt < 0)
			goto data_err;
		transflag = 0;
		minfo.mode=0;
		updateinfo();
		return (0);

	case TYPE_E:
		reply(553, "TYPE E not implemented.");
		transflag = 0;
		return (-1);

	case TYPE_A:
		while ((c = getc(instr)) != EOF) {
			byte_count++;
			if (c == '\n')
				bare_lfs++;
			while (c == '\r') {
				if (ferror(outstr))
					goto data_err;
				if ((c = getc(instr)) != '\n') {
					(void) putc ('\r', outstr);
					if (c == '\0' || c == EOF)
						goto contin2;
				}
			}
			(void) putc(c, outstr);
	contin2:	;
		}
		fflush(outstr);
		if (ferror(instr))
			goto data_err;
		if (ferror(outstr))
			goto file_err;
		transflag = 0;
		if (bare_lfs) {
			lreply(230, "WARNING! %d bare linefeeds received in ASCII mode", bare_lfs);
			printf("   File may not have transferred correctly.\r\n");
		}
		return (0);
	default:
		reply(550, "Unimplemented TYPE %d in receive_data", type);
		transflag = 0;
		return (-1);
	}

data_err:
	transflag = 0;
	perror_reply(426, "Data Connection");
	return (-1);

file_err:
	transflag = 0;
	perror_reply(452, "Error writing file");
	return (-1);
}

void statfilecmd(char *filename)
{
	char line[BUFSIZ];
	FILE *fin;
	int c;

	(void) sprintf(line, "/bin/ls -lgA %s", filename);
	fin = ftpd_popen(line, "r");
	lreply(211, "status of %s:", filename);
	while ((c = getc(fin)) != EOF) {
		if (c == '\n') {
			if (ferror(stdout)){
				perror_reply(421, "control connection");
				(void) ftpd_pclose(fin);
				dologout(1);
				/* NOTREACHED */
			}
			if (ferror(fin)) {
				perror_reply(551, filename);
				(void) ftpd_pclose(fin);
				return;
			}
			(void) putc('\r', stdout);
		}
		(void) putc(c, stdout);
	}
	(void) ftpd_pclose(fin);
	reply(211, "End of Status");
}

void statcmd(void)
{
	struct sockaddr_in *sin;
	u_char *a, *p;

	lreply(211, "%s FTP server status:", hostname);
	printf("     %s\r\n", version);
	printf("     Connected to %s", remotehost);
	if (!isdigit(remotehost[0]))
		printf(" (%s)", inet_ntoa(his_addr.sin_addr));
	printf("\r\n");
	if (logged_in) {
		if (guest)
			printf("     Logged in anonymously\r\n");
		else
			printf("     Logged in as %s\r\n", pw->user_handle);
	} else if (askpasswd)
		printf("     Waiting for password\r\n");
	else
		printf("     Waiting for user name\r\n");
	printf("     TYPE: %s", typenames[type]);
	if (type == TYPE_A || type == TYPE_E)
		printf(", FORM: %s", formnames[form]);
	if (type == TYPE_L)
#if NBBY == 8
		printf(" %d", NBBY);
#else
		printf(" %d", bytesize);	/* need definition! */
#endif
	printf("; STRUcture: %s; transfer MODE: %s\r\n",
	    strunames[stru], modenames[mode]);
	if (data != -1)
		printf("     Data connection open\r\n");
	else if (pdata != -1) {
		printf("     in Passive mode");
		sin = &pasv_addr;
		goto printaddr;
	} else if (usedefault == 0) {
		printf("     PORT");
		sin = &data_dest;
printaddr:
		a = (u_char *) &sin->sin_addr;
		p = (u_char *) &sin->sin_port;
#define UC(b) (((int) b) & 0xff)
		printf(" (%d,%d,%d,%d,%d,%d)\r\n", UC(a[0]),
			UC(a[1]), UC(a[2]), UC(a[3]), UC(p[0]), UC(p[1]));
#undef UC
	} else
		printf("     No data connection\r\n");
	reply(211, "End of status");
}

void fatal(char *s)
{
	reply(451, "Error in server: %s\n", s);
	reply(221, "Closing connection due to server error.");
	dologout(0);
	/* NOTREACHED */
}

void reply(int n, const char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	printf("%d ", n);
	vprintf(fmt, args);
	printf("\r\n");
	fflush(stdout);
	va_end(args);
}

void lreply(int n, const char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	printf("%d- ", n);
	vprintf(fmt, args);
	printf("\r\n");
	fflush(stdout);
	va_end(args);
}

void ack(char *s)
{
	reply(250, "%s command successful.", s);
}

void nack(const char *s)
{
	reply(502, "%s command not implemented.", s);
}

/* ARGSUSED */
void yyerror(char *s)
{
	char *cp;

	if ((cp = strchr(cbuf,'\n')) != 0)
		*cp = '\0';
	reply(500, "'%s': command not understood.", cbuf);
}

void _delete(char *name)
{
	struct stat st;

	if (!checkpath(name)) {
		reply(550,"Illegal filename");
		return;
	}

	if (stat(name, &st) < 0) {
		perror_reply(550, name);
		return;
	}

	if (unlink(name) < 0) {
		perror_reply(550, name);
		return;
	}
	ack("DELE");
}

void cwd(char *path)
{
	if (checkpath(path)) {
		chdir(path);
		if (strstr(path,"ul")) showfile("ftp-upload.txt");
		ack("CWD");
	} else 
		reply(550,"Don't even try it.");
}

void makedir(char *name)
{
	if (!checkpath(name)) {
		reply(550,"Illegal filename");
		return;
	}
	mkdir(name,0755);
}

void removedir(char *name)
{
	if (!checkpath(name)) {
		reply(550,"Illegal filename");
		return;
	}
	rmdir (name);
}

void pwd(void)
{
	char path[MAXPATHLEN + 1];

	if (getcwd(path, sizeof(path)) == (char *) NULL)
		reply(550, "%s.", path);
	else
		reply(257, "\"%s\" is current directory.", path);
}

char *renamefrom(char *name)
{
	struct stat st;
	
	if (!checkpath(name)) {
		reply(550,"Illegal filename");
		return 0;
	}
        if (stat(name, &st) < 0) {
		perror_reply(550, name);
		return ((char *)0);
	}
        reply(350, "File exists, ready for destination name");
	return (name);
}

/* FIXME: 'to' should not contain '/' */
void renamecmd(char *from, char *to)
{
	if (!checkpath(from) || !checkpath(to)) {
		reply(550, "Illegal filename");
		return;
	}
	if (rename(from, to) < 0)
		perror_reply(550, "rename");
	else
		ack("RNTO");
}

void dolog(struct sockaddr_in *sin)
{
	struct hostent *hp = gethostbyaddr((char *)&sin->sin_addr,
		sizeof (struct in_addr), AF_INET);
	time_t t, time();
	extern char *ctime();

	if (hp)
		(void) strncpy(remotehost, hp->h_name, sizeof (remotehost));
	else
		(void) strncpy(remotehost, inet_ntoa(sin->sin_addr),
		    sizeof (remotehost));

	sprintf(proctitle, "%s: connected", remotehost);
	setproctitle(proctitle);

	if (logging) {
		t = time((time_t *) 0);
		syslog(LOG_INFO, "connection from %s at %s",
		    remotehost, ctime(&t));
	}
}

/*
 * Record logout in wtmp file
 * and exit with supplied status.
 */
void dologout(int status)
{
	if (logged_in) {
		(void) seteuid((uid_t)0);
		update_wtmp();
	}
	if (infoname) unlink(infoname);

	/* beware of flushing buffers after a SIGPIPE */
	_exit(status);
}

void myoob(void)
{
	char *cp;

	/* only process if transfer occurring */
	if (!transflag)
		return;
	cp = tmpline;
	if (getline(cp, 7, stdin) == NULL) {
		reply(221, "You could at least say goodbye.");
		dologout(0);
	}
	upper(cp);
	if (strcmp(cp, "ABOR\r\n") == 0) {
		tmpline[0] = '\0';
		reply(426, "Transfer aborted. Data connection closed.");
		reply(226, "Abort successful");
		longjmp(urgcatch, 1);
	}
	if (strcmp(cp, "STAT\r\n") == 0) {
		if (file_size != (off_t) -1)
			reply(213, "Status: %lu of %lu bytes transferred",
			    byte_count, file_size);
		else
			reply(213, "Status: %lu bytes transferred", byte_count);
	}
}

/*
 * Note: a response of 425 is not mentioned as a possible response to
 * 	the PASV command in RFC959. However, it has been blessed as
 * 	a legitimate response by Jon Postel in a telephone conversation
 *	with Rick Adams on 25 Jan 89.
 */
void passive(void)
{
	int len;
	u_short port;
	register char *p, *a;

	if (!logged_in) {
		reply(530, "You are not logged in");
		return;
	}

	pdata = socket(AF_INET, SOCK_STREAM, 0);
	if (pdata < 0) {
		perror_reply(425, "Can't open passive connection");
		return;
	}

	if (restricted_data_ports) {
		for (port = FTP_DATA_BOTTOM; port <= FTP_DATA_TOP; port++) {
			pasv_addr = ctrl_addr;
			pasv_addr.sin_port = htons(port);
			(void) seteuid((uid_t)0);
			if (bind(pdata, (struct sockaddr *)&pasv_addr,
				 sizeof(pasv_addr)) < 0) {
				(void) seteuid((uid_t)mcfg.CFG_BBSUID);
				if (errno == EADDRINUSE)
					continue;
				else
					goto pasv_error;
			}
			(void) seteuid((uid_t)mcfg.CFG_BBSUID);
			break;
		}
		if (port > FTP_DATA_TOP)
			goto pasv_error;
	} else {
		pasv_addr = ctrl_addr;
		pasv_addr.sin_port = 0;
		(void) seteuid((uid_t)0);
		if (bind(pdata, (struct sockaddr *)&pasv_addr,
			 sizeof(pasv_addr)) < 0) {
			(void) seteuid((uid_t)mcfg.CFG_BBSUID);
			goto pasv_error;
		}
		(void) seteuid((uid_t)mcfg.CFG_BBSUID);
	}

	len = sizeof(pasv_addr);
	if (getsockname(pdata, (struct sockaddr *) &pasv_addr, &len) < 0)
		goto pasv_error;
	if (listen(pdata, 1) < 0)
		goto pasv_error;
	a = (char *) &pasv_addr.sin_addr;
	if (forced_passive_ip)
		a = (char *) &passive_ip.sin_addr;
	p = (char *) &pasv_addr.sin_port;

#define UC(b) (((int) b) & 0xff)

	reply(227, "Entering Passive Mode (%d,%d,%d,%d,%d,%d)", UC(a[0]),
		UC(a[1]), UC(a[2]), UC(a[3]), UC(p[0]), UC(p[1]));
	return;

pasv_error:
	(void) close(pdata);
	pdata = -1;
	perror_reply(425, "Can't open passive connection");
	return;
}

/*
 * Generate unique name for file with basename "local".
 * The file named "local" is already known to exist.
 * Generates failure reply on error.
 */
char *gunique(char *local)
{
	static char new[MAXPATHLEN];
	struct stat st;
	char *cp = strrchr(local, '/');
	int count = 0;

	if (cp)
		*cp = '\0';
	if (stat(cp ? local : ".", &st) < 0) {
		perror_reply(553, cp ? local : ".");
		return((char *) 0);
	}
	if (cp)
		*cp = '/';
	(void) strcpy(new, local);
	cp = new + strlen(new);
	*cp++ = '.';
	for (count = 1; count < 100; count++) {
		(void) sprintf(cp, "%d", count);
		if (stat(new, &st) < 0)
			return(new);
	}
	reply(452, "Unique file name cannot be created.");
	return((char *) 0);
}

/*
 * Format and send reply containing system error number.
 */
void perror_reply(int code, char *string)
{
	reply(code, "%s: %s.", string, strerror(errno));
}

static char *onefile[] = {
	"",
	0
};

void send_file_list(char *whichfiles)
{
	struct stat st;
	DIR *dirp = NULL;
	struct dirent *dir;
	FILE *dout = NULL;
	register char **dirlist, *dirname;
	int simple = 0;

	if (strpbrk(whichfiles, "~{[*?") != NULL) {
		extern char **ftpglob(), *globerr;

		globerr = NULL;
		dirlist = ftpglob(whichfiles);
		if (globerr != NULL) {
			reply(550, globerr);
			return;
		} else if (dirlist == NULL) {
			errno = ENOENT;
			perror_reply(550, whichfiles);
			return;
		}
	} else {
		onefile[0] = whichfiles;
		dirlist = onefile;
		simple = 1;
	}

	if (setjmp(urgcatch)) {
		transflag = 0;
		return;
	}
	while ((dirname = *dirlist++) != NULL) {
		if (stat(dirname, &st) < 0) {
			/*
			 * If user typed "ls -l", etc, and the client
			 * used NLST, do what the user meant.
			 */
			if (dirname[0] == '-' && *dirlist == NULL &&
			    transflag == 0) {
				retrieve("/bin/ls %s", dirname);
				return;
			}
			perror_reply(550, whichfiles);
			if (dout != NULL) {
				(void) fclose(dout);
				transflag = 0;
				data = -1;
				pdata = -1;
			}
			return;
		}

		if ((st.st_mode&S_IFMT) == S_IFREG) {
			if (dout == NULL) {
				dout = dataconn("file list", (off_t)-1, "w");
				if (dout == NULL)
					return;
				transflag++;
			}
			fprintf(dout, "%s%s\n", dirname,
				type == TYPE_A ? "\r" : "");
			byte_count += strlen(dirname) + 1;
			continue;
		} else if ((st.st_mode&S_IFMT) != S_IFDIR)
			continue;

		if ((dirp = opendir(dirname)) == NULL)
			continue;

		while ((dir = readdir(dirp)) != NULL) {
			char nbuf[MAXPATHLEN];

#if 0  /* strlen(dir->d_name) is recommended instead of dir->d_namlen */
			if (dir->d_name[0] == '.' && dir->d_namlen == 1)
				continue;
			if (dir->d_name[0] == '.' && dir->d_name[1] == '.' &&
			    dir->d_namlen == 2)
				continue;
#else
			if (strcmp(dir->d_name, ".") == 0)
				continue;
			if (strcmp(dir->d_name, "..") == 0)
				continue;
#endif

			sprintf(nbuf, "%s/%s", dirname, dir->d_name);

			/*
			 * We have to do a stat to insure it's
			 * not a directory or special file.
			 */
			if (simple || (stat(nbuf, &st) == 0 &&
			    (st.st_mode&S_IFMT) == S_IFREG)) {
				if (dout == NULL) {
					dout = dataconn("file list", (off_t)-1,
						"w");
					if (dout == NULL)
						return;
					transflag++;
				}
				if (nbuf[0] == '.' && nbuf[1] == '/')
					fprintf(dout, "%s%s\n", &nbuf[2],
						type == TYPE_A ? "\r" : "");
				else
					fprintf(dout, "%s%s\n", nbuf,
						type == TYPE_A ? "\r" : "");
				byte_count += strlen(nbuf) + 1;
			}
		}
		(void) closedir(dirp);
	}

	if (dout == NULL)
		reply(550, "No files found.");
	else if (ferror(dout) != 0)
		perror_reply(550, "Data connection");
	else
		reply(226, "Transfer complete.");

	transflag = 0;
	if (dout != NULL)
		(void) fclose(dout);
	data = -1;
	pdata = -1;
}

static int cmppasswds(char *passwd, unsigned char *thepw)
{
	MD_CTX context;
	unsigned char digest[16];
	char newpw[30];
	int i;

	strcpy(newpw,passwd);
	strupr(newpw);

        MDInit (&context);
        MDUpdate (&context, newpw, strlen(newpw));
        MDFinal (digest, &context);

	for (i=0;i<16;i++) {
		if(thepw[i]!=digest[i]) return(0);
	}
	return(1);
}

static int dupecheck(const char *name)
{
	struct DayDream_Conference *tconf=confs;
	struct DayDream_MsgBase *tbase;
	
	while(tconf->CONF_NUMBER!=255 && tconf->CONF_NUMBER) {
		int bcnt;
		if (!(tconf->CONF_ATTRIBUTES & (1L<<7))) {
			char de[1024];
			if (findfile(name,de,tconf)) {
				return 1;
			}
		}
		/* FIXME: very suspicious code */
		tbase = (struct DayDream_MsgBase *) tconf + 1;
		bcnt=tconf->CONF_MSGBASES;
		while(bcnt) {
			tbase++;
			bcnt--;
		}
		tconf=(struct DayDream_Conference *)tbase;
	}
	return 0;
}

static int findfile(const char *file, char *de, struct DayDream_Conference *tc)
{
	char buf1[1024];
	char buf2[1024];
	int fd1;
	FILE *plist;
	const char *s;
	DIR *dh;
	struct dirent *dent;
	struct DayDream_Conference *co=tc;
	
	*de=0;
	s=file;
	while(*s) {
		if (*s=='/') return 0;
		s++;
	}	
	sprintf(buf1, "%s/data/paths.dat", co->CONF_PATH);
	if ((plist=fopen(buf1,"r"))) {
		while(fgetsnolf(buf1,512,plist))
		{
			sprintf(buf2,"%s%s",buf1,file);
			if (*buf1 && ((fd1=open(buf2,O_RDONLY)) > -1)) {
				fclose(plist);
				close(fd1);

				strcpy(de,buf2);
				return 1;
			}
		}
		fclose(plist);
	}

	sprintf(buf1, "%s/data/paths.dat", co->CONF_PATH);
	if ((plist=fopen(buf1,"r"))) {
		while(fgetsnolf(buf1,512,plist))
		{
			if (*buf1 && (dh=opendir(buf1)))
			{
				while((dent=readdir(dh)))
				{
					if (!strcmp(dent->d_name,".") || (!strcmp(dent->d_name,".."))) continue;
					if (!strcasecmp(dent->d_name,file)) {
						sprintf(de,"%s%s",buf1,dent->d_name);
						break;
					}
				}
				closedir(dh);
			}
		}
		fclose(plist);
	}
	if (*de) return 1;
	return 0;
}

static char *fgetsnolf(char *buf, int n, FILE *fh)
{
	char *hih;
	char *s;
	
	hih=fgets(buf,n,fh);
	if (!hih) return 0;
	s=buf;
	while (*s)
	{
		if (*s==13 || *s==10) {
			*s=0;
			break;
		}
		s++;
	}
	return hih;
}

static int checkpath(char *pa)
{
	char olddir[1024];
	struct stat st;
	char buf[1024];
	
	getcwd(olddir, sizeof(olddir));
	if ( (stat(pa, &st) <0) || (S_ISDIR(st.st_mode)==0) ) {
		_splitpath(pa,buf,0);
		if (!*buf) return 1;
	} else {
		strcpy(buf,pa);
	}
	if (chdir(buf)>0) return 0;
	getcwd(buf, sizeof(olddir));
	if (!strncmp(buf,myroot,strlen(myroot))) {
		chdir(olddir);
		return 1;
	}
	chdir(olddir);
	return 0;
}

static void _splitpath(char *file, char *path, char *name)
{
	char *s, *t;
	
	s=file;
	while(*s) s++;
	
	while (*s!='/') {
		if (s==file) {
			*path=0;
			if (name) strcpy(name,file);
			return;
		} else {
			s--;
		}
	}
	s++;
	t=file;
	while(t!=s) {
		*path++=*t++;
	}
	*path=0;
	if (name) strcpy(name,t);
}



